"""AUTOGENERATED. DO NOT EDIT."""

from keras.src.optimizers.schedules.learning_rate_schedule import CosineDecay
from keras.src.optimizers.schedules.learning_rate_schedule import CosineDecayRestarts
from keras.src.optimizers.schedules.learning_rate_schedule import ExponentialDecay
from keras.src.optimizers.schedules.learning_rate_schedule import InverseTimeDecay
from keras.src.optimizers.schedules.learning_rate_schedule import LearningRateSchedule
from keras.src.optimizers.schedules.learning_rate_schedule import PiecewiseConstantDecay
from keras.src.optimizers.schedules.learning_rate_schedule import PolynomialDecay
from keras.src.optimizers.schedules.learning_rate_schedule import deserialize
from keras.src.optimizers.schedules.learning_rate_schedule import serialize
