"""AUTOGENERATED. DO NOT EDIT."""

from keras.src.initializers import deserialize
from keras.src.initializers import get
from keras.src.initializers import serialize
from keras.src.initializers.initializers import Initializer
from keras.src.initializers.initializers_v1 import HeNormal as he_normal
from keras.src.initializers.initializers_v1 import HeUniform as he_uniform
from keras.src.initializers.initializers_v1 import LecunNormal as lecun_normal
from keras.src.initializers.initializers_v1 import LecunUniform as lecun_uniform
from keras.src.initializers.initializers_v1 import RandomNormal
from keras.src.initializers.initializers_v1 import RandomNormal as normal
from keras.src.initializers.initializers_v1 import RandomNormal as random_normal
from keras.src.initializers.initializers_v1 import RandomUniform
from keras.src.initializers.initializers_v1 import RandomUniform as random_uniform
from keras.src.initializers.initializers_v1 import RandomUniform as uniform
from keras.src.initializers.initializers_v1 import TruncatedNormal
from keras.src.initializers.initializers_v1 import TruncatedNormal as truncated_normal
from tensorflow.python.ops.init_ops import Constant
from tensorflow.python.ops.init_ops import Constant as constant
from tensorflow.python.ops.init_ops import GlorotNormal as glorot_normal
from tensorflow.python.ops.init_ops import GlorotUniform as glorot_uniform
from tensorflow.python.ops.init_ops import Identity
from tensorflow.python.ops.init_ops import Identity as identity
from tensorflow.python.ops.init_ops import Ones
from tensorflow.python.ops.init_ops import Ones as ones
from tensorflow.python.ops.init_ops import Orthogonal
from tensorflow.python.ops.init_ops import Orthogonal as orthogonal
from tensorflow.python.ops.init_ops import VarianceScaling
from tensorflow.python.ops.init_ops import Zeros
from tensorflow.python.ops.init_ops import Zeros as zeros
